/*
 *This file is part of MlView
 *
 *MlView is free software; you can redistribute it and/or modify it under the terms of 
 *the GNU General Public License as published by the Free Software Foundation; either version 2, 
 *or (at your option) any later version.
 *
 *MlView is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 *without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *See the GNU General Public License for more details.
 *
 *You should have received a copy of the GNU General Public License along with MlView; 
 *see the file COPYING. If not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 *Copyright 2001-2002 Dodji Seketeli.
 */

#include "mlview-file-descriptor.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/unistd.h>
#include <fcntl.h>
#include <libxml/uri.h>
#include <libxml/nanohttp.h>
#include <libxml/nanoftp.h>


/**
 *@file
 *The class definition of #MlViewFileDescriptor.
 */

/**
 *The private members of #MlViewFileDescriptor class.
 */
struct _MlViewFileDescriptorPrivate
{
	/**the string representation of the uri*/
	gchar *uri_str ;
	
	/**
	 *The URI of this file descriptor.
	 *Note that the xmlURI type comes from
	 *libxml2. To get the libxml2 soft, go
	 *to http://xmlsoft.org
	 */
	xmlURI * uri; 
	
	/**
	 *The last modification date of the current file.
	 *is valid only in the case of a local file
	 */
	time_t last_modif_date ;
	
        /**
	 *wether the current file is local
	 *or not.
	 */
	gboolean is_local ;
};

enum MLVIEW_FILE_DESCRIPTOR_STATUS {
	MLVIEW_FILE_DESCRIPTOR_BAD_PARAM = -2,
	MLVIEW_FILE_DESCRIPTOR_FETCH_FAILED = -1,
	MLVIEW_FILE_DESCRIPTOR_OK = 0,	
	MLVIEW_FILE_DESCRIPTOR_IS_LOCAL = 1	
} ;

#define PRIVATE(file_desc) (file_desc->private)

/**
 *The constructor of #MlViewFileDescriptor class.
 *
 *@param a_file_uri The uri to associate to the file descriptor.
 *This null terminated char is duplicated. Then caller still owns
 *the pointer.
 *Note that this argument _must_ be a uri.
 *
 *@return the newly created file descriptor or NULL if could not
 *Have been created.
 */
MlViewFileDescriptor *
mlview_file_descriptor_new (const gchar * a_file_uri)
{
	MlViewFileDescriptor * fd = NULL ;
	struct stat file_stats ;
	int status ;

	g_return_val_if_fail (a_file_uri != NULL , NULL) ;
	fd  = g_malloc0 (sizeof(MlViewFileDescriptor)) ;
	PRIVATE (fd) = g_malloc (sizeof(MlViewFileDescriptorPrivate)) ;

	if( !(PRIVATE (fd)->uri = xmlParseURI (a_file_uri)) ){
		g_free (PRIVATE (fd)) ;
		g_free (fd) ;
		return NULL ;
	}

	PRIVATE (fd)->uri_str = g_strdup (a_file_uri) ;
						
	PRIVATE (fd)->is_local = (!PRIVATE (fd)->uri->scheme || !strcmp (PRIVATE(fd)->uri->scheme, "file")) ? TRUE : FALSE ;

	if (PRIVATE(fd)->is_local == TRUE && PRIVATE(fd)->uri && PRIVATE(fd)->uri->path) {
		if ( (status = stat (PRIVATE (fd)->uri->path, &file_stats)) == 0) {
			PRIVATE (fd)->last_modif_date = file_stats.st_mtime ;
		} else {
			PRIVATE(fd)->last_modif_date = 0 ;
		}
	}

	return fd ;
}


/**
 *The destructor of the #MlViewFileDescriptor class.
 *@param a_file_desc the instance of #MlViewFileDescriptor to destroy.
 *
 */
void 
mlview_file_descriptor_destroy (MlViewFileDescriptor *a_file_desc)
{
	
	g_return_if_fail (a_file_desc != NULL) ;

	if (!PRIVATE (a_file_desc))
		return ;

	if (PRIVATE (a_file_desc)->uri && PRIVATE (a_file_desc)->uri->path != NULL) {
		xmlFreeURI (PRIVATE (a_file_desc)->uri) ;
		PRIVATE (a_file_desc)->uri = NULL ;
	}
	if (PRIVATE (a_file_desc)->uri_str) {
		g_free (PRIVATE (a_file_desc)->uri_str) ;
		PRIVATE (a_file_desc)->uri_str = NULL ;
	}

	g_free (PRIVATE (a_file_desc)) ;
	PRIVATE (a_file_desc) = NULL ;
	g_free (a_file_desc) ;
	a_file_desc = NULL ;
}


/**
 *Sets the last modified time of the current instance 
 *of file descriptor to the current time. 
 *
 *@param a_file_desc the current instance of MlViewFileDescriptor.
 */
void
mlview_file_descriptor_update_modified_time (MlViewFileDescriptor *a_file_desc)
{
	g_return_if_fail (a_file_desc != NULL) ;
	g_return_if_fail (PRIVATE (a_file_desc) != NULL) ;

	if (PRIVATE (a_file_desc)->is_local)
		PRIVATE (a_file_desc)->last_modif_date = time (NULL) ;
}


/**
 *Check wether the in memory file has been modified.
 *
 *@param a_file_desc the current instance of #MlViewFileDescriptor. (in memory file)
 *@param a_is_modified out parameter. Is set to TRUE if the lastp modif date of
 *the in memory file is newer than the last modif date of the on disk file. This parameter
 *is valid if and only if the return value of this method is 0.
 *
 *@return 0 if the operation went well.
 */
gint
mlview_file_descriptor_is_modified (const MlViewFileDescriptor *a_file_desc,
				    gboolean * a_is_modified)
{
	struct stat file_stats ;
	int status ;

	*a_is_modified = FALSE ;
	g_return_val_if_fail (a_file_desc != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_file_desc) != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_file_desc)->uri, -1) ;

	if (PRIVATE (a_file_desc)->uri->path == NULL || !PRIVATE (a_file_desc)->is_local)
		return -1 ;

	status = stat (PRIVATE (a_file_desc)->uri->path, & file_stats) ;

	if (status != 0)
		return -1 ;

	/*
	 *if the last modif date  of the on memory file is 
	 *newer than the last modif date of the on disk file, 
	 *return TRUE
	 */
	*a_is_modified = (file_stats.st_mtime < PRIVATE (a_file_desc)->last_modif_date) ? TRUE : FALSE ;

	return 0 ;
}


/**
 *Tells if 
 *@param a_file_desc the current instance of MlViewFileDescriptor
 *
 */
gint
mlview_file_descriptor_is_readable (const MlViewFileDescriptor *a_file_desc, 
				    gboolean *a_is_readable)
{
	struct stat file_stats ;
	int status ;
	mode_t mode ;
	
	*a_is_readable = FALSE ;

	g_return_val_if_fail ( a_file_desc != NULL, -1) ;
	g_return_val_if_fail ( PRIVATE (a_file_desc) != NULL, -1) ;
	g_return_val_if_fail ( PRIVATE (a_file_desc)->uri != NULL, -1) ;

	if (PRIVATE (a_file_desc)->uri->path == NULL)
		return -1 ;
	
	if (PRIVATE (a_file_desc)->is_local){
		status = stat (PRIVATE (a_file_desc)->uri->path, &file_stats) ;
		if (status != 0)
			return -1 ;
		mode = file_stats.st_mode ;
	
		/*If this file offers the read access to the current user, return TRUE*/
		*a_is_readable = (((file_stats.st_mode & S_IRUSR) == S_IRUSR))?TRUE:FALSE ;
	}
	return 0 ;
}


/**
 *
 *@param a_file_desc the current instance of MlViewFileDescriptor
 *
 *
 */
gint
mlview_file_descriptor_is_writeable (const MlViewFileDescriptor *a_file_desc, gboolean *a_is_writeable)
{
	struct stat file_stats ;
	int status ;
	mode_t mode ;
	
	*a_is_writeable = FALSE ;

	g_return_val_if_fail( a_file_desc != NULL, -1) ;
	g_return_val_if_fail( PRIVATE (a_file_desc) != NULL, -1) ;
	g_return_val_if_fail( PRIVATE (a_file_desc)->uri != NULL, -1) ;

	if (PRIVATE (a_file_desc)->uri->path == NULL)
		return -1 ;

	if (PRIVATE (a_file_desc)->is_local) {
		status = stat (PRIVATE (a_file_desc)->uri->path, &file_stats) ;
		
		if (status != 0)
			return -1 ;

		mode = file_stats.st_mode ;
	
		/*If this file offers the write access to the current user, return TRUE*/
		*a_is_writeable = (((file_stats.st_mode & S_IWUSR) == S_IWUSR))?TRUE :FALSE ;
	}
	return 0 ;
	
}

/**
 *
 *@param a_file_desc The current instance of file descriptor.
 *
 *
 *@return NULL. The file path can also be NULL. 
 */
gchar * 
mlview_file_descriptor_get_file_path(const MlViewFileDescriptor *a_file_desc)
{
	g_return_val_if_fail (a_file_desc != NULL, NULL) ;
	g_return_val_if_fail (PRIVATE (a_file_desc) != NULL, NULL) ;
	g_return_val_if_fail (PRIVATE (a_file_desc)->uri != NULL, NULL) ;

	return PRIVATE (a_file_desc)->uri->path ;
}

/**
 *
 *@param a_file_path the file path to set.
 *@param a_file_desc the current instance of MlViewFileDescriptor.
 *
 *
 */
void 
mlview_file_descriptor_set_file_path (MlViewFileDescriptor *a_file_desc, gchar *a_file_path)
{
	struct stat file_stats ;
	int status ;
	
	g_return_if_fail (a_file_desc != NULL) ;
	g_return_if_fail (PRIVATE (a_file_desc) != NULL) ;
	g_return_if_fail (PRIVATE (a_file_desc)->uri) ;

	PRIVATE (a_file_desc)->uri->path = a_file_path ;
	if (PRIVATE (a_file_desc)->is_local) {
		status = stat (a_file_path, &file_stats) ;

		if (status == 0)
			PRIVATE (a_file_desc)->last_modif_date = file_stats.st_mtime ;
		PRIVATE (a_file_desc)->last_modif_date = 0 ;
	}
}


/**
 *
 *@param a_file_desc the current instance of MlViewFileDescriptor
 *
 *
 *@return the last modif date of the file pointed by file desc. 
 */
guint 
mlview_file_descriptor_get_lmd (const MlViewFileDescriptor *a_file_desc)
{
	g_return_val_if_fail (a_file_desc != NULL, 0) ;
	g_return_val_if_fail (PRIVATE (a_file_desc) != NULL, 0) ;

	return PRIVATE (a_file_desc)->last_modif_date ;
}


/**
 *Sets the last modified date to a_last_modif time 
 *
 *@param a_file_desc the current instance of MlViewFileDescriptor.
 *@param a_last_modif_time the new modif time.
 */
void 
mlview_file_descriptor_set_lmd (MlViewFileDescriptor *a_file_desc,
				guint a_last_modif_time)
{
	g_return_if_fail (a_file_desc != NULL) ;
	g_return_if_fail (PRIVATE (a_file_desc) != NULL) ;

	PRIVATE (a_file_desc)->last_modif_date = a_last_modif_time ;
}


/**
 *
 *@param a_is_dir response. Equals TRUE if the file descriptor points to a directory and false if not.
 *@param a_file_desc the current file descriptor
 *
 *
 */
gint 
mlview_file_descriptor_is_dir (const MlViewFileDescriptor *a_file_desc, gboolean * a_is_dir)
{
	struct stat file_stats ;
	gint status ;

	*a_is_dir=FALSE ;
	g_return_val_if_fail (a_file_desc != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_file_desc) != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_file_desc)->uri != NULL, -1) ;

	if (PRIVATE (a_file_desc)->uri->path == NULL || !PRIVATE(a_file_desc)->is_local)
		return FALSE ;
	status = stat (PRIVATE (a_file_desc)->uri->path, &file_stats) ;
	
	if (status != 0)
		return -1 ;
	*a_is_dir = (S_ISDIR(file_stats.st_mode))?TRUE:FALSE ;	
	return 0 ;
}

/**
 *
 */
gint
mlview_file_descriptor_is_regular_file (const MlViewFileDescriptor *a_file_desc, gboolean * a_is_reg)
{
	gint status ;
	struct stat file_stats ;

	*a_is_reg=FALSE ;
	g_return_val_if_fail (a_file_desc != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_file_desc) != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_file_desc)->uri != NULL, -1) ;

	if (PRIVATE (a_file_desc)->uri->path == NULL || !PRIVATE (a_file_desc)->is_local)
		return FALSE ;
	status = stat (PRIVATE (a_file_desc)->uri->path, &file_stats) ;
	
	if (status != 0)
		return -1 ;
	*a_is_reg = (S_ISREG(file_stats.st_mode))?TRUE:FALSE ;
	return 0 ;
}


/**
 *
 */
gint
mlview_file_descriptor_create_directory (MlViewFileDescriptor *a_file_desc, guint a_mode)
{
	g_return_val_if_fail (a_file_desc != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_file_desc) != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_file_desc)->is_local == TRUE, -1) ;
	g_return_val_if_fail (PRIVATE (a_file_desc)->uri != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_file_desc)->uri->path != NULL, -1) ;	

	return mkdir (PRIVATE (a_file_desc)->uri->path, a_mode) ;
}

/**
 *
 *@param a_mode the file mode
 *@param a_file_desc creates a file descriptor.
 *
 *
 */
gint 
mlview_file_descriptor_create_file (MlViewFileDescriptor *a_file_desc, guint a_mode)
{
	gint fd=0 ;

	g_return_val_if_fail (a_file_desc != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_file_desc) != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_file_desc)->is_local == TRUE, -1) ;
	g_return_val_if_fail (PRIVATE (a_file_desc)->uri != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_file_desc)->uri->path != NULL, -1) ;
	
	fd = open (PRIVATE (a_file_desc)->uri->path, O_CREAT , a_mode) ;

	if (fd != -1) {/*open succeed*/
		close (fd) ;
		return 0 ;
	}
	return -1 ;

	/*return open (PRIVATE (a_file_desc)->uri->path, O_CREAT , a_mode) ;*/

}


gint
mlview_file_descriptor_is_local (MlViewFileDescriptor * a_file_desc, gboolean * a_is_local)
{
	*a_is_local = FALSE ;
	g_return_val_if_fail (a_file_desc != NULL, -1) ;
	g_return_val_if_fail (PRIVATE(a_file_desc) != NULL, -1) ;

	*a_is_local = PRIVATE (a_file_desc)->is_local ;
	return 0 ;
}


/**
 *If the file is remote, and the url is valid, fetches the file using the
 *the protocol suggested by the url scheme (http or ftp) and stores the result
 *on the local disk at the file path designated by file a_file_path. 
 *
 *@param a_file_path the local file path where to store the fetched file.
 *@param a_file_desc the current file descriptor.
 *
 *
 *@return the status. 
 */
enum MLVIEW_FILE_DESCRIPTOR_STATUS
mlview_file_descriptor_fetch_file (MlViewFileDescriptor * a_file_desc,
				   const gchar * a_file_path) {
	gboolean is_local = TRUE ;
	gint status = 0 ;

	g_return_val_if_fail (a_file_desc != NULL, MLVIEW_FILE_DESCRIPTOR_BAD_PARAM) ;
	g_return_val_if_fail (PRIVATE (a_file_desc) != NULL, MLVIEW_FILE_DESCRIPTOR_BAD_PARAM) ;

	if (mlview_file_descriptor_is_local (a_file_desc, &is_local))
		return MLVIEW_FILE_DESCRIPTOR_BAD_PARAM ;

	if (is_local == TRUE)
		return MLVIEW_FILE_DESCRIPTOR_IS_LOCAL ;

	if (!PRIVATE (a_file_desc)->uri_str) 
		return MLVIEW_FILE_DESCRIPTOR_BAD_PARAM ;

	status = xmlNanoHTTPFetch (PRIVATE (a_file_desc)->uri_str, a_file_path, NULL) ;

	if (status == -1)
		return MLVIEW_FILE_DESCRIPTOR_FETCH_FAILED ;

	return MLVIEW_FILE_DESCRIPTOR_OK ;
}
