/*
 *This file is part of MlView
 *
 *MlView is free software; you can redistribute it and/or modify it under the terms of 
 *the GNU General Public License as published by the Free Software Foundation; either version 2, 
 *or (at your option) any later version.
 *
 *MlView is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 *without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *See the GNU General Public License for more details.
 *
 *You should have received a copy of the GNU General Public License along with MlView; 
 *see the file COPYING. If not, write to the Free Software Foundation, 
 *Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 *Copyright 2001-2002 Dodji Seketeli.
 */

#include <libxml/tree.h>
#include <libxml/valid.h>

#include "mlview-xml-document.h"
#include "mlview-parsing-utils.h"
#include "mlview-settings-management.h"
#include "mlview-file-selection.h"
#include "mlview-file-descriptor.h"
#include "mlview-global-settings.h"
#include "utils.h"

/*FIXME: remove all the references to ->tree_view.
 *Make sure to transform then in wrappers of the daniel veillard's libxml.
 *That way, all the real actions made on the underlying xml data structure will
 *be made through this object. This object will then emit signals to
 *notify it's state to all the views that are in relation with it. Of course, this
 *object should not know anything about the views that holds it ...
 *
 *Damn, i think we gonna turn this gnome-mlview into a nice piece of software. :)
 */
struct _MlViewXMLDocumentPrivate
{
	MlViewFileDescriptor *file_desc ;
	/*MlViewXMLDocTreeView * tree_view ;*/

	/*the underlying xml doc data structure*/
	xmlDocPtr xml_doc ;

	/*the current xml application context*/
	MlViewAppContext * app_context ;
};

enum {
	DOCUMENT_CHANGED = 0 ,
	NODE_CUT,
	CHILD_NODE_ADDED,
	PREV_SIBLING_NODE_INSERTED,
	NEXT_SIBLING_NODE_INSERTED,
	CONTENT_CHANGED,
	NAME_CHANGED,
	NODE_CHANGED,
	FILE_PATH_CHANGED,
	NUMBER_OF_SIGNALS
} ;

#define PRIVATE(mlview_xml_doc) (mlview_xml_doc->private)
#define MLVIEW_XML_DOCUMENT_DEFAULT_VALIDATION_ON FALSE ;


static void 
mlview_xml_document_do_interactive_validation_if_needed (MlViewXMLDocument * a_doc) ;

static gint 
mlview_xml_document_validate_xml_doc_against_a_dtd (MlViewXMLDocument * a_doc, 
						    MlViewExternalSubsetDefinition * 
						    a_subset_def) ;

static void mlview_xml_document_destroy (GtkObject *a_xml_doc) ;

#define MLVIEW_XML_DOCUMENT_NODE_CLIPBOARD_SIZE 128
static GtkObjectClass *gv_parent_class = NULL ;
static guint gv_mlview_xml_document_signals [NUMBER_OF_SIGNALS] = {0} ;
static xmlNode * gv_clipboard[MLVIEW_XML_DOCUMENT_NODE_CLIPBOARD_SIZE] = {0} ;
static guint gv_clipboard_index = 0 ;

/*======================================================
 *private helper functions
 *=====================================================*/

/**
 *
 */
static void
mlview_xml_document_do_interactive_validation_if_needed (MlViewXMLDocument * a_doc)
{
	gchar * validation_is_on = NULL ;

	g_return_if_fail(a_doc != NULL) ;
	g_return_if_fail(PRIVATE(a_doc) != NULL) ;
	g_return_if_fail(PRIVATE(a_doc)->xml_doc != NULL) ;

	if (PRIVATE (a_doc)->app_context)
		validation_is_on = 
			mlview_app_context_get_settings_value (PRIVATE (a_doc)->app_context,
							       MLVIEW_STG_K_IS_VALIDATION_ON) ;
	
	if (validation_is_on
	    && (strcmp (validation_is_on, MLVIEW_STG_V_YES) == 0)) {
		
		mlview_xml_document_validate (a_doc) ;
	}	    	   
}



/*
 *mlview_xml_document_validate_xml_doc_against_a_dtd:
 *@a_doc:The document to validate
 *@a_subset_def: a description of the DTD.
 *
 *Loads the DTD pointed by a_subset_def and validate it.
 *Note that if the doc contains a dtd already, that dtd is overloaded by the new one
 *without being freed.
 *
 *FIXME: this function does not support catalog/remote dtd so far. It only support dtds that
 *are on the local disk.
 *
 *Returns 0 if validation went well, >0 if validation failed, <0 if could not even start validation.
 *The possible return values are:
 *-3 one of the parameter is not correct.
 * 0 the document is valid against the Dtd
 * 1 the document is not valid against the dtd.
 * 2 parsing of the dtd failed.
 */
static gint
mlview_xml_document_validate_xml_doc_against_a_dtd (MlViewXMLDocument * a_doc,
						    MlViewExternalSubsetDefinition * a_subset_def)
{
	gint validity_status=1 ;

	g_return_val_if_fail(a_doc != NULL, -3) ;
	g_return_val_if_fail(PRIVATE(a_doc) != NULL, -3) ;
	g_return_val_if_fail(PRIVATE(a_doc)->xml_doc != NULL, -3) ;
	g_return_val_if_fail(a_subset_def != NULL, -3) ;

	if ( !mlview_xml_document_associate_dtd(a_doc, a_subset_def) ) 
		validity_status = mlview_xml_document_validate(a_doc) ;
	else
		validity_status = 2; /*association (probably dtd parsing) failed*/

	return validity_status ;
}

/***********************************************
 *private gtk framework methods						
 ***********************************************/
static void
mlview_xml_document_init_class (MlViewXMLDocumentClass * a_klass)
{
	GtkObjectClass * object_class = NULL ;

	g_return_if_fail (a_klass != NULL) ;
	g_return_if_fail (MLVIEW_IS_XML_DOCUMENT_CLASS (a_klass)) ;

	object_class = GTK_OBJECT_CLASS (a_klass) ;
	gv_parent_class = gtk_type_class (GTK_TYPE_OBJECT) ;
	object_class->destroy = mlview_xml_document_destroy ;

	/*signal creation code*/
	gv_mlview_xml_document_signals[DOCUMENT_CHANGED] = 
		gtk_signal_new ("document-changed",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (MlViewXMLDocumentClass,
						   document_changed),
				gtk_marshal_NONE__NONE,
				GTK_TYPE_NONE, 0) ;

	gv_mlview_xml_document_signals[NODE_CUT] = 
		gtk_signal_new ("node-cut",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (MlViewXMLDocumentClass, node_cut),
				gtk_marshal_NONE__POINTER_POINTER,
				GTK_TYPE_NONE, 2, 
				GTK_TYPE_POINTER, GTK_TYPE_POINTER) ;

	gv_mlview_xml_document_signals [CHILD_NODE_ADDED] = 
		gtk_signal_new ("child-node-added",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (MlViewXMLDocumentClass, child_node_added),
				gtk_marshal_NONE__POINTER_POINTER,
				GTK_TYPE_NONE, 2, 
				GTK_TYPE_POINTER, GTK_TYPE_POINTER) ;

	gv_mlview_xml_document_signals [PREV_SIBLING_NODE_INSERTED] = 
		gtk_signal_new ("prev-sibling-node-inserted",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (MlViewXMLDocumentClass, prev_sibling_node_inserted),
				gtk_marshal_NONE__POINTER_POINTER,
				GTK_TYPE_NONE, 2, 
				GTK_TYPE_POINTER, GTK_TYPE_POINTER) ;

	gv_mlview_xml_document_signals [NEXT_SIBLING_NODE_INSERTED] = 
		gtk_signal_new ("next-sibling-node-inserted",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (MlViewXMLDocumentClass, next_sibling_node_inserted),
				gtk_marshal_NONE__POINTER_POINTER,
				GTK_TYPE_NONE, 2, 
				GTK_TYPE_POINTER, GTK_TYPE_POINTER) ;

	gv_mlview_xml_document_signals [CONTENT_CHANGED] = 
		gtk_signal_new ("content-changed",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (MlViewXMLDocumentClass, content_changed),
				gtk_marshal_NONE__POINTER,
				GTK_TYPE_NONE, 1, 
				GTK_TYPE_POINTER) ;

	gv_mlview_xml_document_signals [NAME_CHANGED] = 
		gtk_signal_new ("name-changed",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (MlViewXMLDocumentClass, 
						   name_changed),
				gtk_marshal_NONE__POINTER,
				GTK_TYPE_NONE, 1, 
				GTK_TYPE_POINTER) ;

	gv_mlview_xml_document_signals [NODE_CHANGED] = 
		gtk_signal_new ("node-changed",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (MlViewXMLDocumentClass, 
						   node_changed),
				gtk_marshal_NONE__POINTER,
				GTK_TYPE_NONE, 1, 
				GTK_TYPE_POINTER) ;

	gv_mlview_xml_document_signals [FILE_PATH_CHANGED] =
		gtk_signal_new ("file-path-changed",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (MlViewXMLDocumentClass, 
						   file_path_changed),
				gtk_marshal_NONE__NONE,
				GTK_TYPE_NONE, 0) ;

	gtk_object_class_add_signals (object_class, 
				      gv_mlview_xml_document_signals, 
				      NUMBER_OF_SIGNALS) ;

	a_klass->document_changed = NULL ;
	a_klass->node_cut = NULL ;
	a_klass->child_node_added = NULL ;
	a_klass->node_changed = NULL ;
	a_klass->content_changed = NULL ;
	a_klass->name_changed = NULL ;
}


static void
mlview_xml_document_init (MlViewXMLDocument * a_xml_doc)

{
	g_return_if_fail (a_xml_doc != NULL) ;
	g_return_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_doc)) ;

	PRIVATE (a_xml_doc) = g_malloc0(sizeof(MlViewXMLDocumentPrivate)) ;
}

/*=============================================
 *public methods
 *=============================================*/

guint 
mlview_xml_document_get_type (void)
{
	static guint mlview_xml_document_type = 0 ;
	
	if (!mlview_xml_document_type) {
		static const GtkTypeInfo t_info = {
			"MlViewXMLDocument",
			sizeof (MlViewXMLDocument),
			sizeof (MlViewXMLDocumentClass) ,
			(GtkClassInitFunc) mlview_xml_document_init_class,
			(GtkObjectInitFunc) mlview_xml_document_init,
			NULL, NULL, NULL
		} ;
		mlview_xml_document_type = gtk_type_unique (GTK_TYPE_OBJECT, &t_info) ;
	}
	return mlview_xml_document_type ;
}


/**
 *
 *@param a_file_path the file path of the xml document. Must be NULL terminated string. If NULL, the internal file_path is set to NULL. if NOT NULL, the string is duplicated.
 *so that the caller can free it. 
 *
 *
 */
MlViewXMLDocument *
mlview_xml_document_new (xmlDocPtr a_xml_doc, MlViewAppContext *a_context)
{
	MlViewXMLDocument *result = NULL ;
	xmlDocPtr xml_doc=NULL ;
	xmlNodePtr xml_node ;

	result = gtk_type_new (MLVIEW_TYPE_XML_DOCUMENT) ;

	xml_doc = a_xml_doc ;

	if (xml_doc == NULL) {
		xml_doc = xmlNewDoc ("1.0") ;
		xml_node = xmlNewNode (NULL, "") ;
		xmlDocSetRootElement (xml_doc, xml_node) ;
	}

	PRIVATE (result)->xml_doc = xml_doc ;
	PRIVATE (result)->app_context = a_context ;
	return result ;
}


/**
 *
 *@param a_doc the current xml document.
 *@param a_context the current application context to set to this component.
 *
 *
 */
void
mlview_xml_document_set_app_context (MlViewXMLDocument * a_doc, 
				     MlViewAppContext * a_context)
{
	g_return_if_fail (a_doc != NULL) ;
	g_return_if_fail (PRIVATE (a_doc) != NULL) ;
	g_return_if_fail (a_context != NULL) ;
	g_return_if_fail (MLVIEW_IS_APP_CONTEXT (a_context)) ;


	PRIVATE (a_doc)->app_context = a_context ;
}


/**
 *
 *@param a_doc the current xml document.
 *
 *
 */
MlViewAppContext *
mlview_xml_document_get_app_context (MlViewXMLDocument * a_doc) 
{
	g_return_val_if_fail (a_doc, NULL) ;
	g_return_val_if_fail (PRIVATE (a_doc) != NULL, NULL) ;
	
	return PRIVATE (a_doc)->app_context ;
}


/************************************
 *clipboard methods.
 ***********************************/

/**
 *Stores the a copy of the xmlNode given in argument into the node clipboard.
 *The cilpboard is a circular buffer. It cannot be overflowed. 
 *
 *@param a_visual_node the pointer to xmlNode to store in the clipboard. If it is a NULL pointer, a NULL pointer is actually stored into clipboard.
 *
 *
 */
void
mlview_xml_document_copy_node_to_clipboard (xmlNode *a_xml_node, 
					   xmlDoc *a_doc)
{
	g_return_if_fail (a_xml_node != NULL) ;

	g_return_if_fail (sizeof (gv_clipboard) / sizeof (xmlNodePtr) > 0) ;

	if (gv_clipboard_index >= sizeof (gv_clipboard) / sizeof (xmlNodePtr))
		gv_clipboard_index = 0 ;
	gv_clipboard[gv_clipboard_index] = xmlDocCopyNode (a_xml_node, a_doc , TRUE) ;
	gv_clipboard_index ++ ;
}


/**
 *
 *@param a_xml_doc the destination document. Is usefull to know how to handle namespace copy stuffs.
 *
 *
 *@return a different copy of the same xmlNode that has been previously stored. 
 */
xmlNode *
mlview_xml_document_get_node_from_clipboard (xmlDoc * a_xml_doc)
{
	xmlNodePtr xml_node=NULL ;

	g_return_val_if_fail(sizeof(gv_clipboard)/sizeof(xmlNodePtr) > 0, NULL) ;
	if(gv_clipboard_index >= sizeof(gv_clipboard)/sizeof(xmlNodePtr))
		gv_clipboard_index = sizeof(gv_clipboard)/sizeof(xmlNodePtr) - 1 ;

	xml_node = gv_clipboard[gv_clipboard_index - 1] ;
	if(xml_node != NULL)
		xml_node = xmlDocCopyNode (xml_node, a_xml_doc, 1) ;/*make a recursive copy of the node*/
	return xml_node ;
}


/*************************************
 *Other miscellaneous methods
 *************************************/

/**
 *Opens the xml file contains in the filename given in parameter. 
 *
 *@param a_file_name the total file path of the file that contains the xml document to open
 *@param a_mlview_xml_doc the current instance of MlViewXMLDocument.
 *
 *
 */
MlViewXMLDocument *
mlview_xml_document_open (gchar * a_file_name, MlViewAppContext * a_context)
{
	MlViewXMLDocument *result = NULL;

	xmlDoc * xml_doc ;

	xml_doc = mlview_parsing_utils_load_xml_file (a_file_name, a_context) ;
	g_return_val_if_fail (xml_doc != NULL , NULL) ;

	result = mlview_xml_document_new (xml_doc, a_context) ;

	mlview_xml_document_do_interactive_validation_if_needed (result) ;	
	mlview_xml_document_set_file_path (result, a_file_name) ;

	return result;
}

/**
 *Associates a Dtd to the current xml document. If the document already contains
 *a Dtd, this functions displays a dialog box to ask the user if he wants to overide
 *that Dtd. Depending on the answer of the user, this function does the association
 *or not. 
 *
 *@param a_doc the current mlview xml document.
 *
 *
 *@return 0 if the association has been done, 1 if the user cancelled the operations, 
 */
gint
mlview_xml_document_associate_dtd_interactive (MlViewXMLDocument * a_doc)
{
	MlViewExternalSubsetDefinition * external_subset_def ;

	g_return_val_if_fail (a_doc != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_doc) != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_doc)->app_context != NULL, -1) ;

	external_subset_def = 
		mlview_parsing_utils_let_user_choose_a_dtd (PRIVATE (a_doc)->app_context, 
							    _("Choose a Dtd to associate to the document")) ;
	if ( external_subset_def ) {
		return mlview_xml_document_associate_dtd(a_doc,
							 external_subset_def) ;		
	} else {
		return 1 ;
	}
}


/**
 *Associates a the dtd designated by a_ext_subs_def to the current xml document.
 *Note that if the document contains is already associated to a dtd, the function asks
 *the user if she wants to overide it. If the user agrees, the functions frees the former
 *dtd and makes the new association. If not, nothing is done. 
 *
 *@return 0 if the association has been done, 1 if the user cancelled the operations, 
 */
gint
mlview_xml_document_associate_dtd ( MlViewXMLDocument * a_doc, 
				    MlViewExternalSubsetDefinition * a_ext_subs_def)
{
	xmlDtd * dtd=NULL ;
	gboolean user_cancels=FALSE ;

	g_return_val_if_fail(a_doc != NULL, -1) ;
	g_return_val_if_fail(PRIVATE(a_doc) != NULL, -1) ;

	/*if an external subset already exists, ask the user if she wants to overide it*/
	if ( PRIVATE(a_doc)->xml_doc && PRIVATE(a_doc)->xml_doc->extSubset ) {
		GnomeDialog * dialog ;
		GtkWidget * label ;
		gint button ;

		dialog = GNOME_DIALOG(gnome_dialog_new(_("Document has already has a Dtd"),
						       _("YES"),_("CANCEL"), NULL)) ;
		label = gtk_label_new(_("This Document already has an associated Dtd.\nDo you want really want to associate another Dtd to this document ?")) ;
		gtk_box_pack_start(GTK_BOX(dialog->vbox),
				   label, TRUE,FALSE,0) ;
		gtk_widget_show_all(dialog->vbox) ;
		gnome_dialog_close_hides(dialog, FALSE) ;
		gnome_dialog_set_close (dialog, TRUE) ;
		button = gnome_dialog_run(dialog) ;

		switch ( button ) {
		case 0:/*YES button*/
			break ;
		default :/*CANCEL Button, dialog box closed etc..*/
			user_cancels = TRUE ;
			break ;
		}
	}/*end if*/
	
	if ( user_cancels == TRUE )
		return 1 ;/*user cancelled operations*/

	/*if the user did not cancel the operations, let's do the association*/
	
	/*let's parse the external subset. 
	 *If parsing is OK, we associate that dtd to
	 *the document.
	 */
	if ( a_ext_subs_def->system_id  
	     &&  strcmp(a_ext_subs_def->system_id, "")  
	     &&  (dtd = mlview_parsing_utils_load_a_dtd (a_ext_subs_def, PRIVATE(a_doc)->app_context)) ) {
		
		if (PRIVATE(a_doc)->xml_doc->extSubset) 
			xmlFreeDtd(PRIVATE(a_doc)->xml_doc->extSubset) ;
		PRIVATE(a_doc)->xml_doc->extSubset = dtd ;
		PRIVATE(a_doc)->xml_doc->standalone = 0 ;
		return 0 ; /*dtd parsing OK, association OK*/
	}

	return 2 ;/*dtd parsing failed*/
}


/**
 *Validates the document agains the dtd that is associated to it. 
 *
 *@param a_doc the document to validate.
 *
 *
 */
gint
mlview_xml_document_validate (MlViewXMLDocument * a_doc)
{
	gint validity_status=-2 ;/*defaul value means "there is no dtd to parse"*/

	g_return_val_if_fail (a_doc != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_doc) != NULL, -1) ;
	g_return_val_if_fail (PRIVATE (a_doc)->xml_doc, -1) ;

	if ( PRIVATE(a_doc)->xml_doc->extSubset
	     || PRIVATE(a_doc)->xml_doc->intSubset)
		validity_status =
			mlview_parsing_utils_validate_dtd (PRIVATE(a_doc)->xml_doc,
							   PRIVATE(a_doc)->xml_doc->extSubset,
							   PRIVATE(a_doc)->app_context) ;
	else {
		mlview_app_context_warning (PRIVATE (a_doc)->app_context, _("No DTD was provided. Could not validate the document"))  ;
		return validity_status ;
	}

	/*If the document is valid, tell the user.*/
	if (! validity_status )
		mlview_app_context_message (PRIVATE (a_doc)->app_context, _("The Document is valid :-)")) ;
	else
		mlview_app_context_warning (PRIVATE (a_doc)->app_context, _("The Document is not valid !"))  ;

	return validity_status ;
}


/**
 *
 *@param a_doc the current instance of MlViewXMLDocument
 *
 *
 *@return the number of bytes saved or -1 in case of error. 
 */
gint
mlview_xml_document_save (MlViewXMLDocument * a_doc)
{
	MlViewFileDescriptor * file_desc ;
	gchar * file_path ;
	xmlDocPtr xml_doc ;

	g_return_val_if_fail(a_doc != NULL, -1) ;
	g_return_val_if_fail(PRIVATE(a_doc) != NULL, -1) ;

	xml_doc = PRIVATE(a_doc)->xml_doc ;
	g_return_val_if_fail(xml_doc != NULL, -1) ;

	file_desc = mlview_xml_document_get_file_descriptor(a_doc) ;
	g_return_val_if_fail(file_desc != NULL, -1) ;

	file_path = mlview_file_descriptor_get_file_path(file_desc) ;
	g_return_val_if_fail(file_path != NULL, -1) ;

	return mlview_parsing_utils_save_xml_doc (xml_doc, file_path, PRIVATE(a_doc)->app_context) ;
}


/**
 *
 *@param a_file_path 
 *@param a_doc 
 *
 *
 *@return the number of bytes saved or -1 in case of error. 
 */
gint
mlview_xml_document_save_as (MlViewXMLDocument *a_doc, gchar * a_file_path)
{
	gint result ;
	gchar * old_path ;
	g_return_val_if_fail (a_file_path != NULL, -1) ;

	/*store the old file path in case the save fails.*/
	old_path = g_strdup (mlview_xml_document_get_file_path (a_doc)) ;

	mlview_xml_document_set_file_path (a_doc, a_file_path) ;	

	if ( (result = mlview_xml_document_save (a_doc)) <= 0 )
		mlview_xml_document_set_file_path (a_doc, old_path) ;
	g_free (old_path) ;

	return result ;
}


/**
 *Clones the current instance of MlViewXMLDocument. 
 *
 *@param a_original the current instance of MlViewXMLDocument.
 *
 *
 */
MlViewXMLDocument*
mlview_xml_document_clone (MlViewXMLDocument *a_original)
{	 
	MlViewXMLDocument * result ;
	MlViewFileDescriptor * file_desc ;
	gchar * file_path ;

	if ( a_original == NULL )
		return NULL ;
	file_desc = mlview_xml_document_get_file_descriptor (a_original) ;
	file_path = mlview_file_descriptor_get_file_path (file_desc);
	result = mlview_xml_document_open (file_path, PRIVATE (a_original)->app_context) ;
	return result ;
}

/**
 *
 *@param a_xml_doc 
 *@param a_file_path 
 *
 *
 */
void
mlview_xml_document_set_file_path (MlViewXMLDocument *a_xml_doc,
				   gchar *a_file_path)
{
	g_return_if_fail (a_xml_doc != NULL) ;
	g_return_if_fail (PRIVATE (a_xml_doc) != NULL) ;

	if( PRIVATE (a_xml_doc)->file_desc == NULL ) {
		PRIVATE (a_xml_doc)->file_desc =
			mlview_file_descriptor_new (a_file_path) ;
	} else {
		mlview_file_descriptor_set_file_path (PRIVATE (a_xml_doc)->file_desc,
						      a_file_path) ;
	}

	gtk_signal_emit (GTK_OBJECT (a_xml_doc),
			 gv_mlview_xml_document_signals [FILE_PATH_CHANGED]) ;
}


/**
 *
 *@param a_xml_doc 
 *
 *
 */
gchar *
mlview_xml_document_get_file_path (MlViewXMLDocument *a_xml_doc)
{
	g_return_val_if_fail (a_xml_doc != NULL, NULL) ;
	g_return_val_if_fail (PRIVATE(a_xml_doc) != NULL, NULL) ;

	return mlview_file_descriptor_get_file_path (PRIVATE (a_xml_doc)->file_desc) ;
}


/**
 *Adds a child node to the a_parent_node node. 
 *
 *@param a_xml_node the node to add.
 *@param a_xml_doc the current instance of MlViewXMLDocument.
 *@param a_parent_node the parent of the node to add.
 *
 *
 */
xmlNode *
mlview_xml_document_add_child_node (MlViewXMLDocument *a_xml_doc,
				    xmlNode *a_parent_xml_node,
				    xmlNode *a_xml_node, gboolean a_subtree_required)
{
	xmlNode * result = NULL  ;
	gchar *  validation_is_on = NULL ;

	g_return_val_if_fail (a_xml_doc != NULL, NULL) ;
	g_return_val_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_doc), NULL) ;
	g_return_val_if_fail (PRIVATE (a_xml_doc) != NULL, NULL) ;
	g_return_val_if_fail (a_parent_xml_node != NULL, NULL) ;
	g_return_val_if_fail ((a_parent_xml_node->type == XML_ELEMENT_NODE) 
			      || (a_parent_xml_node->type == XML_DOCUMENT_NODE), 
			      NULL) ;
	g_return_val_if_fail (a_xml_node != NULL, NULL) ;

	result = xmlAddChild (a_parent_xml_node, a_xml_node) ;
	g_return_val_if_fail (result != NULL, NULL) ;

	if (a_subtree_required == TRUE
	    && PRIVATE (a_xml_doc)->app_context
	    && mlview_app_context_settings_exist (PRIVATE (a_xml_doc)->app_context)) {
		validation_is_on = 
			mlview_app_context_get_settings_value (PRIVATE (a_xml_doc)->app_context,
							       MLVIEW_STG_K_IS_VALIDATION_ON) ;
		if (validation_is_on && !strcmp (validation_is_on, MLVIEW_STG_V_YES)) {

			enum MLVIEW_PARSING_UTILS_STATUS  status ;
			status =
				mlview_parsing_utils_build_required_attributes_list (PRIVATE (a_xml_doc)->app_context, 
										     result) ;
			status =
				mlview_parsing_utils_build_required_children_tree (PRIVATE (a_xml_doc)->app_context,
										   &result) ;
		}

	}

	/*emit the proper signals*/
	gtk_signal_emit (GTK_OBJECT (a_xml_doc),
			 gv_mlview_xml_document_signals [CHILD_NODE_ADDED],
			 a_parent_xml_node, a_xml_node) ;

	gtk_signal_emit (GTK_OBJECT (a_xml_doc),
			 gv_mlview_xml_document_signals [DOCUMENT_CHANGED]) ;
	return result ;
}


/**
 *Cuts the node a_xml_node. The node is stored in the node clipboard
 *and then unlinked from the xml tree. 
 *
 *@param a_xml_doc the current xml node
 *
 *
 *@return the xmlNode* cut or NULL if something bad happened. 
 */
xmlNode *
mlview_xml_document_cut_node (MlViewXMLDocument * a_xml_doc, xmlNode * a_xml_node)
{
	xmlNode * parent_node = NULL , *result = NULL;

	g_return_val_if_fail (a_xml_doc != NULL, NULL) ;
	g_return_val_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_doc), NULL) ;
	g_return_val_if_fail (PRIVATE (a_xml_doc) != NULL, NULL) ;
	g_return_val_if_fail (PRIVATE (a_xml_doc)->xml_doc != NULL, NULL) ;
	g_return_val_if_fail (a_xml_node != NULL, NULL) ;

	g_return_val_if_fail (a_xml_node->doc == PRIVATE (a_xml_doc)->xml_doc, NULL) ;

	parent_node = a_xml_node->parent ;

	g_return_val_if_fail (parent_node != NULL, NULL) ;

	mlview_xml_document_copy_node_to_clipboard (a_xml_node, PRIVATE (a_xml_doc)->xml_doc) ;
	xmlUnlinkNode (a_xml_node) ;
	result = a_xml_node ;

	gtk_signal_emit (GTK_OBJECT (a_xml_doc),
			 gv_mlview_xml_document_signals [NODE_CUT],
			 parent_node, a_xml_node) ;

	gtk_signal_emit (GTK_OBJECT (a_xml_doc),
			 gv_mlview_xml_document_signals [DOCUMENT_CHANGED]) ;
	
	return result ;
}


/**
 *Sets a new content a_content to a_node. 
 *
 *@param a_content the new content to set to a_node.
 *@param a_xml_document the current xml document.
 *@param a_node the node to set the new content to.
 *
 *
 *@return the a_node with the new content or NULL if something bad happened. 
 */
xmlNode *
mlview_xml_document_set_node_content (MlViewXMLDocument * a_xml_document,
				      xmlNode *a_node, gchar * a_content)
{
	g_return_val_if_fail (a_xml_document != NULL, NULL) ;
	g_return_val_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_document), NULL) ;
	g_return_val_if_fail (a_node != NULL, NULL) ;
	
	xmlNodeSetContent (a_node, a_content) ;

	gtk_signal_emit (GTK_OBJECT (a_xml_document), 
			 gv_mlview_xml_document_signals [CONTENT_CHANGED], a_node) ;

	gtk_signal_emit (GTK_OBJECT (a_xml_document), 
			 gv_mlview_xml_document_signals [NODE_CHANGED], a_node) ;

	gtk_signal_emit (GTK_OBJECT (a_xml_document), 
			 gv_mlview_xml_document_signals [DOCUMENT_CHANGED]) ;

	return a_node ;
}

/**
 *Sets the name of a_node to a_name. 
 *
 *@param a_xml_document the current xml document.
 *@param a_name the new name to set.
 *@param a_node the node to set the name on.
 *
 *
 *@return a_node with it new name or NULL if something bad happened. 
 */
xmlNode *
mlview_xml_document_set_node_name (MlViewXMLDocument * a_xml_document,
				   xmlNode *a_node, gchar * a_name)
{
	g_return_val_if_fail (a_xml_document != NULL, NULL) ;
	g_return_val_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_document), NULL) ;
	g_return_val_if_fail (a_node != NULL, NULL) ;

	xmlNodeSetName (a_node, a_name) ;

	gtk_signal_emit (GTK_OBJECT (a_xml_document), 
			 gv_mlview_xml_document_signals [NAME_CHANGED], a_node) ;
 
	gtk_signal_emit (GTK_OBJECT (a_xml_document), 
			 gv_mlview_xml_document_signals [NODE_CHANGED], a_node) ;

	gtk_signal_emit (GTK_OBJECT (a_xml_document), 
			 gv_mlview_xml_document_signals [DOCUMENT_CHANGED]) ;
	return a_node ;
}


/**
 *Inserts a_sibling_node as the previous sibling node of a_xml_node. 
 *
 *@param a_xml_node the current reference node. Must be not NULL.
 *@param a_xml_document the current xml document. Must be not NULL.
 *@param a_sibling_node a the new sibling node to insert. Must be not NULL.
 *
 *
 *@return the newly insterted sibling (a_sibling_node) or NULL if something BAD happened. 
 */
xmlNode *
mlview_xml_document_insert_prev_sibling_node (MlViewXMLDocument *a_xml_document,
					      xmlNode *a_sibling_node, xmlNode *a_xml_node,
					      gboolean a_subtree_required)
{
	xmlNode * result = NULL ;
	gchar *  validation_is_on = NULL ;

	g_return_val_if_fail (a_xml_document != NULL, NULL) ;
	g_return_val_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_document), NULL) ;
	g_return_val_if_fail (PRIVATE (a_xml_document) != NULL, NULL) ;
	g_return_val_if_fail (a_sibling_node != NULL, NULL) ;
	g_return_val_if_fail (a_xml_node != NULL, NULL) ;

	result = xmlAddPrevSibling (a_sibling_node, a_xml_node) ;

	if (a_subtree_required == TRUE
	    && PRIVATE (a_xml_document)->app_context
	    && mlview_app_context_settings_exist (PRIVATE (a_xml_document)->app_context)) {
		validation_is_on =
			mlview_app_context_get_settings_value (PRIVATE (a_xml_document)->app_context,
							       MLVIEW_STG_K_IS_VALIDATION_ON) ;

		if (validation_is_on && !strcmp (validation_is_on, MLVIEW_STG_V_YES)) {

			enum MLVIEW_PARSING_UTILS_STATUS  status ;
			status = 
				mlview_parsing_utils_build_required_attributes_list (PRIVATE (a_xml_document)->app_context, 
										     result) ;
			status = 
				mlview_parsing_utils_build_required_children_tree (PRIVATE (a_xml_document)->app_context, 
										   &result) ;
		}
			
	}

	gtk_signal_emit (GTK_OBJECT (a_xml_document),
			 gv_mlview_xml_document_signals [PREV_SIBLING_NODE_INSERTED],
			 a_sibling_node,
			 a_xml_node) ;

	gtk_signal_emit (GTK_OBJECT (a_xml_document),
			 gv_mlview_xml_document_signals [DOCUMENT_CHANGED]) ;

	return result ;
}


/**
 *Inserts a_sibling_node as  the next sibling of a_xml_node. 
 *
 *@param a_xml_node the reference xml node.
 *@param a_xml_document the current xml document.
 *@param a_sibling_node the new sibling node to insert.
 *
 *
 *@return the newly inserted next sibling node (a_sibling_node) or 
 */
xmlNode *
mlview_xml_document_insert_next_sibling_node (MlViewXMLDocument *a_xml_document,
					      xmlNode *a_sibling_node, xmlNode *a_xml_node,
					      gboolean a_subtree_required)
{
	xmlNode * result = NULL ;
	gchar *  validation_is_on = NULL ;

	g_return_val_if_fail (a_xml_document != NULL, NULL) ;
	g_return_val_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_document), NULL) ;
	g_return_val_if_fail (PRIVATE (a_xml_document) != NULL, NULL) ;
	g_return_val_if_fail (a_sibling_node != NULL, NULL) ;
	g_return_val_if_fail (a_xml_node != NULL, NULL) ;
	
	result = xmlAddNextSibling (a_sibling_node, a_xml_node) ;
	
	if (a_subtree_required == TRUE
	    && PRIVATE (a_xml_document)->app_context
	    && mlview_app_context_settings_exist (PRIVATE (a_xml_document)->app_context)) {
		validation_is_on =
			mlview_app_context_get_settings_value (PRIVATE (a_xml_document)->app_context,
							       MLVIEW_STG_K_IS_VALIDATION_ON) ;

		if (validation_is_on && !strcmp (validation_is_on, MLVIEW_STG_V_YES)) {

			enum MLVIEW_PARSING_UTILS_STATUS  status ;

			status =
				mlview_parsing_utils_build_required_attributes_list (PRIVATE (a_xml_document)->app_context, 
										      result) ;
			status = 
				mlview_parsing_utils_build_required_children_tree (PRIVATE (a_xml_document)->app_context, 
										   &result) ;
		}
			
	}

	gtk_signal_emit (GTK_OBJECT (a_xml_document),
			 gv_mlview_xml_document_signals [NEXT_SIBLING_NODE_INSERTED],
			 a_sibling_node,
			 a_xml_node) ;

	gtk_signal_emit (GTK_OBJECT (a_xml_document),
			 gv_mlview_xml_document_signals [DOCUMENT_CHANGED]) ;

	return result ;
}


void
mlview_xml_document_paste_node_as_child (MlViewXMLDocument *a_xml_document,
					 xmlNode *a_parent_node)
{
	xmlNode * xml_node = NULL ;

	g_return_if_fail (a_xml_document != NULL) ;
	g_return_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_document)) ;
	g_return_if_fail (PRIVATE (a_xml_document)) ;
	g_return_if_fail (a_parent_node != NULL) ;

	xml_node = mlview_xml_document_get_node_from_clipboard (PRIVATE (a_xml_document)->xml_doc) ;
	g_return_if_fail (xml_node != NULL) ;

	mlview_xml_document_add_child_node (a_xml_document, a_parent_node, 
					    xml_node, FALSE) ;
}


void
mlview_xml_document_paste_node_as_sibling (MlViewXMLDocument * a_xml_document, xmlNode *a_parent_node,
					   xmlNode * a_sibling_node, gboolean a_previous)
{
	xmlNode *xml_node = NULL ;

	g_return_if_fail (a_xml_document != NULL) ;
	g_return_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_document)) ;
	g_return_if_fail (PRIVATE (a_xml_document) != NULL) ;
	g_return_if_fail (a_parent_node != NULL) ;
	g_return_if_fail (a_sibling_node != NULL) ;

	xml_node = mlview_xml_document_get_node_from_clipboard (PRIVATE (a_xml_document)->xml_doc) ;
	g_return_if_fail (xml_node != NULL) ;

	if (a_previous == FALSE) {
		mlview_xml_document_insert_next_sibling_node (a_xml_document,
							      a_sibling_node, xml_node, FALSE) ;
	} else {
		mlview_xml_document_insert_prev_sibling_node (a_xml_document,
							      a_sibling_node, xml_node, FALSE) ;
	}
}


/**
 *
 *@param a_doc the current instance of MlViewXMLDocument.
 *
 *
 *@return FALSE. So, please, make sure 
 */
gboolean
mlview_xml_document_needs_saving (MlViewXMLDocument *a_doc)
{
	MlViewFileDescriptor *file_desc ;
	gboolean is_modified ;

	g_return_val_if_fail (a_doc != NULL, FALSE) ;
	g_return_val_if_fail (PRIVATE(a_doc) != NULL, FALSE) ;
	
	file_desc = mlview_xml_document_get_file_descriptor (a_doc) ;
	if ( file_desc == NULL )
		return TRUE ;
	
	if ( mlview_file_descriptor_is_modified (file_desc, &is_modified) )
		return FALSE ;
	return is_modified ;
}

/**
 *
 */
void 
mlview_xml_document_destroy (GtkObject *a_xml_doc)
{
	MlViewXMLDocument * xml_doc = NULL ;

	g_return_if_fail (a_xml_doc != NULL) ;
	g_return_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_doc)) ;
	
	xml_doc = MLVIEW_XML_DOCUMENT (a_xml_doc) ;

	if ( PRIVATE (xml_doc) == NULL )
		return ;
	if ( PRIVATE(xml_doc)->file_desc != NULL ) {
		mlview_file_descriptor_destroy (PRIVATE (xml_doc)->file_desc) ;
		PRIVATE (xml_doc)->file_desc = NULL ;
	}
	
	if ( PRIVATE (xml_doc)->xml_doc != NULL ){
		xmlFreeDoc (PRIVATE (xml_doc)->xml_doc) ;
		PRIVATE (xml_doc)->xml_doc = NULL ;
	}

	g_free (PRIVATE (xml_doc)) ;
	PRIVATE (xml_doc) = NULL ;

	if (GTK_OBJECT_CLASS (gv_parent_class)->destroy) {
		GTK_OBJECT_CLASS (gv_parent_class)->destroy (a_xml_doc) ;
	}
}

/**
 *
 *@param a_xml_doc 
 *
 *
 */
MlViewFileDescriptor *
mlview_xml_document_get_file_descriptor (MlViewXMLDocument *a_xml_doc)
{
	g_return_val_if_fail (a_xml_doc != NULL, NULL) ;
	g_return_val_if_fail (PRIVATE (a_xml_doc) != NULL, NULL) ;
	
	return PRIVATE (a_xml_doc)->file_desc ;
}

/**
 *
 *@param a_xml_doc 
 *@param a_file_desc 
 *
 *
 */
void 
mlview_xml_document_set_file_descriptor (MlViewXMLDocument *a_xml_doc, MlViewFileDescriptor *a_file_desc)
{
	g_return_if_fail (a_xml_doc != NULL) ;
	g_return_if_fail (PRIVATE(a_xml_doc) != NULL) ;
	
	PRIVATE (a_xml_doc)->file_desc = a_file_desc ;
}

/**
 *
 *@param a_xml_doc 
 *
 *
 */
xmlDocPtr
mlview_xml_document_get_xml_document (MlViewXMLDocument *a_xml_doc)
{
	g_return_val_if_fail (a_xml_doc != NULL, NULL) ;
	g_return_val_if_fail (PRIVATE(a_xml_doc) != NULL, NULL) ;
	
	return PRIVATE (a_xml_doc)->xml_doc ;
}


/**
 *
 *@param a_mem_doc 
 *@param a_doc 
 *
 *
 */
void
mlview_xml_document_set_xml_document (MlViewXMLDocument *a_doc, xmlDocPtr a_mem_doc)
{
	g_return_if_fail (a_doc != NULL) ;
	g_return_if_fail (PRIVATE(a_doc) != NULL) ;

	PRIVATE (a_doc)->xml_doc = a_mem_doc ;	
}
